import unittest
import itertools
from collections import deque
from dgim import Dgim


class TestDgim(unittest.TestCase):
    def test_get_count(self):
        dgim = Dgim(12)
        stream = iter([
            False, False, True, False, True, True, True, False, True,
            True, False, False, True, False, True, True, False
        ])
        for elt in stream:
            dgim.update(elt)
        self.assertEquals(7, dgim.get_count())

    def test_get_count_without_update(self):
        """Example from chapter 4 of "Mining of Massing Datasets"""
        crt_timestamp = 65
        # hand built dgim
        queues = [
            deque([crt_timestamp - 1, crt_timestamp - 2]),
            deque([crt_timestamp - 4]),
            deque([crt_timestamp - 8]),
            deque()
        ]

        dgim = Dgim(10)
        dgim.timestamp = crt_timestamp
        dgim.queues = queues
        self.assertEquals(6, dgim.get_count())

    def test_count_empty_stream(self):
        dgim = Dgim(10)
        self.assertEqual(0, dgim.get_count())

    def test_N_is_null(self):
        dgim = Dgim(0)
        stream = iter([True, False, False, True])
        for elt in stream:
            dgim.update(elt)
        self.assertEquals(0, dgim.get_count())

    def test_N_is_one(self):
        dgim = Dgim(1)
        dgim.update(True)
        self.assertEqual(1, dgim.get_count())
        dgim.update(False)
        self.assertEqual(0, dgim.get_count())

    def test_N_is_two(self):
        dgim = Dgim(2)
        dgim.update(True)
        self.assertEqual(1, dgim.get_count())
        dgim.update(True)
        self.assertEqual(2, dgim.get_count())
        dgim.update(True)
        self.assertEqual(2, dgim.get_count())

    def test_bucket_drop(self):
        crt_timestamp = 65
        # hand built dgim
        queues = [
            deque([crt_timestamp - 1, crt_timestamp - 2]),
            deque([crt_timestamp - 4]),
            deque()
        ]
        dgim = Dgim(6)
        dgim.timestamp = crt_timestamp
        dgim.queues = queues
        dgim.oldest_bucket_timestamp = crt_timestamp - 4
        self.assertEquals(3, dgim.nb_buckets)
        dgim.update(0)
        self.assertEquals(3, dgim.nb_buckets)
        dgim.update(0)
        self.assertEquals(2, dgim.nb_buckets)

    def test_only_zeros(self):
        dgim = Dgim(10)
        for elt in itertools.repeat(False, 1000):
            dgim.update(elt)
            self.assertEqual(0, dgim.get_count())

    def test_r_computation(self):
        dgim = Dgim(10, 0.5)
        self.assertEqual(2, dgim.r)

        dgim = Dgim(10, 0.1)
        self.assertEqual(10, dgim.r)

    def test_invalid_error_rates(self):
        self.assertRaises(ValueError, Dgim, 10, 0)
        self.assertRaises(ValueError, Dgim, 10, -0.1)
        self.assertRaises(ValueError, Dgim, 10, 1.1)

    def test_is_bucket_too_old(self):
        dgim = Dgim(10)
        dgim.timestamp = 15
        self.assertFalse(dgim.is_bucket_too_old(6))
        self.assertTrue(dgim.is_bucket_too_old(5))
        self.assertTrue(dgim.is_bucket_too_old(16))

        dgim.timestamp = 5
        self.assertFalse(dgim.is_bucket_too_old(16))
        self.assertTrue(dgim.is_bucket_too_old(15))



import random


def generate_random_stream(length):
    """Generate a random stream of booleans.
    :param length: the stream length
    :type length: int
    :returns: iterator
    """
    for i in range(length):
        yield bool(random.randint(0, 1))



import math
from collections import deque

class Dgim(object):
    """An implementation of the DGIM algorithm.
    It estimates the number of "True" present the last N elements
    of a boolean stream.
    The datastructure it uses is very compact and
    has a memory complexity of O(log(N)^2).

    The algorithm is described in:
    Datar, Mayur, et al. "Maintaining stream statistics over sliding windows."
    SIAM Journal on Computing 31.6 (2002): 1794-1813.

    An explanation of the algorithm can also be found here:
    http://infolab.stanford.edu/~ullman/mmds/ch4.pdf
    """

    def __init__(self, N, error_rate=0.5):
        """Constructor
        :param N: size of the sliding window
        :type N: int
        :param error_rate: the maximum error made by the algorithm.
        The error rate is in ]0, 1]
        Let c be the true result and e the estimate returned by the dgim
        algorithm.
        abs(c-e) < error_rate * c
        :type error_rate: float
        """
        self.N = N

        if not (0 < error_rate <= 1):
            raise ValueError(
                    "Invalid value for error_rate: {}. Error rate should be in ]0, 1].".format(error_rate))

        self.error_rate = error_rate

        #the maximum number of buckets of the same size
        self.r = math.ceil(1/error_rate)
        self.r = max(self.r, 2)

        # the datastructure to store the buckets
        # it is an array of queues
        # queue[i] stores the timestamp of the bucket of size 2^i
        # in descending order
        # this structure makes it easy to :
        # - know how many bucket of the same size there is
        # - update a queue of bucket
        self.queues = []
        if N == 0:
            max_index = -1
        else:
            max_index = int(math.ceil(math.log(N)/math.log(2)))

        for i in range(max_index + 1):
            self.queues.append(deque())

        self.timestamp = 0
        self.oldest_bucket_timestamp = -1  # No bucket so far

    @property
    def nb_buckets(self):
        """Returns the number of buckets.
        :returns: int
        """
        result = 0
        for queue in self.queues:
            result += len(queue)
        return result

    def update(self, elt):
        """Update the stream with one element.
        :param elt: the latest element of the stream
        :type elt: bool
        """
        if self.N == 0:
            return
        self.timestamp = (self.timestamp + 1) % (2 * self.N)
        #check if oldest bucket should be removed
        if self.oldest_bucket_timestamp >= 0 and self.is_bucket_too_old(self.oldest_bucket_timestamp):
            self.drop_oldest_bucket()
        if elt is not True:
            #nothing to do
            return

        carry_over = self.timestamp
        if self.oldest_bucket_timestamp == -1:
            self.oldest_bucket_timestamp = self.timestamp
        for queue in self.queues:
            queue.appendleft(carry_over)
            if len(queue) <= self.r:
                break
            last = queue.pop()
            second_last = queue.pop()
            # merge last two buckets.
            carry_over = second_last
            if last == self.oldest_bucket_timestamp:
                self.oldest_bucket_timestamp = second_last

    def is_bucket_too_old(self, bucket_timestamp):
        """Check if a bucket is too old and should be dropped.
        ;param bucket_timestamp: the bucket timestamp
        :type bucket_timestamp: int
        :returns: bool
        """
        # the buckets are stored modulo 2 * N
        return (self.timestamp - bucket_timestamp) % (2 * self.N) >= self.N

    def drop_oldest_bucket(self):
        """Drop oldest bucket timestamp."""
        for queue in reversed(self.queues):
            if len(queue) > 0:
                queue.pop()
                break
        #update oldest bucket timestamp
        self.oldest_bucket_timestamp = -1
        for queue in reversed(self.queues):
            if len(queue) > 0:
                self.oldest_bucket_timestamp = queue[-1]
                break

    def get_count(self):
        """Returns an estimate of the number of "True"
        in the last N elements of the stream.
        :returns: int
        """
        result = 0
        max_value = 0
        power_of_two = 1
        for queue in self.queues:
            queue_length = len(queue)
            if queue_length > 0:
                max_value = power_of_two
                result += queue_length * power_of_two
            power_of_two = power_of_two << 1
        result -= math.floor(max_value/2)
        return int(result)



import unittest
import itertools
from dgim import Dgim
from dgim.utils import generate_random_stream


class ExactAlgorithm(object):
    """Exact algorithm to count the number of "True"
    in the last N elements of a boolean stream."""
    def __init__(self, N):
        """Constructor
        :param N: size of the sliding window
        :type N: int
        """
        self.N = N
        self.sliding_window = []

    def update(self, elt):
        """Update the stream with one element.
        :param elt: the latest element of the stream
        :type elt: bool
        """
        self.sliding_window.append(elt)
        if len(self.sliding_window) > self.N:
            self.sliding_window.pop(0)

    def get_count(self):
        """Returns an estimate of the number of "True"
        in the last N elements of the stream.
        :returns: int
        """
        return sum(self.sliding_window)


class TestDgimQuality(unittest.TestCase):

    def check_quality_settings(self, dgim, stream):
        """Compare the result "e" returned by dgim with the exact result
        "c" on a stream which elements are 0 or 1.
        The test fails if the dgim result "e" is not in the expected bounds.
        0.5 * c <= e <= 1.5 * c

        :param dgim: the Dgim instance to test
        :type dgim: Dgim
        :param stream: the stream to use. It should contains only 0 or 1 as elements.
        :type stream: iterator
        """
        exact_algorithm = ExactAlgorithm(dgim.N)
        for elt in stream:
            dgim.update(elt)
            exact_algorithm.update(elt)

            exact_result = exact_algorithm.get_count()
            error = abs(dgim.get_count() - exact_result)
            self.assertTrue(error <= dgim.error_rate * exact_result)

    def test_nominal_case(self):
        dgim = Dgim(N=100, error_rate=0.5)
        stream = generate_random_stream(length=10000)
        self.check_quality_settings(dgim=dgim, stream=stream)

    def test_large_N(self):
        dgim = Dgim(N=10000, error_rate=0.5)
        stream = generate_random_stream(length=2000)
        self.check_quality_settings(dgim=dgim, stream=stream)

    def test_short_stream(self):
        dgim = Dgim(N=1000, error_rate=0.5)
        # stream is shorter than N
        stream = generate_random_stream(length=100)
        self.check_quality_settings(dgim=dgim, stream=stream)

    def test_N_is_one(self):
        dgim = Dgim(N=1, error_rate=0.5)
        stream = generate_random_stream(length=10)
        self.check_quality_settings(dgim=dgim, stream=stream)

    def test_N_is_two(self):
        dgim = Dgim(N=2, error_rate=0.5)
        stream = generate_random_stream(length=100)
        self.check_quality_settings(dgim=dgim, stream=stream)

    def test_low_error_rate_case(self):
        dgim = Dgim(N=100, error_rate=0.01)
        stream = generate_random_stream(length=1000)
        self.check_quality_settings(dgim=dgim, stream=stream)

    def test_only_true_case(self):
        dgim = Dgim(N=100, error_rate=0.5)
        stream = itertools.repeat(True, 10000)
        self.check_quality_settings(dgim=dgim, stream=stream)
